/* global after before describe it expect */

import "../src/plugin";
const tagName = "dark-mode-switch";

const assertElementIsInLightMode = () => {
  const testElement = document.body.querySelector(tagName);
  expect(testElement.getCurrentTheme()).to.eq("light");
  expect(testElement.getNextTheme()).to.eq("dark");
  const actualMessage = testElement
    .querySelector("span.theme-switcher-message")
    .innerHTML;
  expect(actualMessage).to.eq("Switch to dark mode");
};

const assertElementIsInDarkMode = () => {
  const testElement = document.body.querySelector(tagName);
  expect(testElement.getCurrentTheme()).to.eq("dark");
  expect(testElement.getNextTheme()).to.eq("light");
  const actualMessage = testElement
    .querySelector("span.theme-switcher-message")
    .innerHTML;
  expect(actualMessage).to.eq("Switch to light mode");
};

const assertBodyCssTheme = (present, absent) => {
  expect(document.body.classList.contains(present)).to.eq(true);
  expect(document.body.classList.contains(absent)).to.eq(false);
};

describe("The dark-mode-switch element", () => {

  it("should be defined as custom element", () => {
    expect(customElements.get(tagName)).not.to.eq(undefined);
  });

  describe("should handle having no default theme", () => {
    before(() => {
      const testElement = document.createElement(tagName);
      document.body.appendChild(testElement);
    });

    it("without updating the attribute", () => {
      const testElement = document.body.querySelector(tagName);
      expect(testElement.getAttribute("theme")).to.eq(null);
    });

    it("render as a light element", () => {
      assertElementIsInLightMode();
    });

    it("update the body's theme to light", () => {
      assertBodyCssTheme("light", "dark");
    });

    after(() => {
      const testElement = document.body.querySelector(tagName);
      document.body.removeChild(testElement);
    });
  });

  describe("should handle using a light theme", () => {
    before(() => {
      const testElement = document.createElement(tagName);
      testElement.setAttribute("theme", "light");
      document.body.appendChild(testElement);
    });

    it("render as a light element", () => {
      assertElementIsInLightMode();
    });

    it("update the body's theme to light", () => {
      assertBodyCssTheme("light", "dark");
    });

    after(() => {
      const testElement = document.body.querySelector(tagName);
      document.body.removeChild(testElement);
    });
  });

  describe("should handle using a dark theme", () => {
    before(() => {
      const testElement = document.createElement(tagName);
      testElement.setAttribute("theme", "dark");
      document.body.appendChild(testElement);
    });

    it("render as a dark element", () => {
      assertElementIsInDarkMode();
    });

    it("update the body's theme to dark", () => {
      assertBodyCssTheme("dark", "light");
    });

    after(() => {
      const testElement = document.body.querySelector(tagName);
      document.body.removeChild(testElement);
    });
  });

  describe("should handle switching between dark and light mode", () => {
    before(() => {
      const testElement = document.createElement(tagName);
      testElement.setAttribute("theme", "dark");
      document.body.appendChild(testElement);
      testElement.click();
    });

    it("switch the attribute to light", () => {
      const testElement = document.body.querySelector(tagName);
      const themeAttribute = testElement.getAttribute("theme");
      expect(themeAttribute).to.eq("light");
    });

    it("render as a light element", () => {
      assertElementIsInLightMode();
    });

    it("update the body's theme to light", () => {
      assertBodyCssTheme("light", "dark");
    });

    after(() => {
      const testElement = document.body.querySelector(tagName);
      document.body.removeChild(testElement);
    });
  });

  describe("should cycle between dark and light mode", () => {
    before(() => {
      const testElement = document.createElement(tagName);
      testElement.setAttribute("theme", "dark");
      document.body.appendChild(testElement);
      testElement.click();
      testElement.click();
    });

    it("switch the attribute to dark", () => {
      const testElement = document.body.querySelector(tagName);
      const themeAttribute = testElement.getAttribute("theme");
      expect(themeAttribute).to.eq("dark");
    });

    it("render as a dark element", () => {
      assertElementIsInDarkMode();
    });

    it("update the body's theme to dark", () => {
      assertBodyCssTheme("dark", "light");
    });

    after(() => {
      const testElement = document.body.querySelector(tagName);
      document.body.removeChild(testElement);
    });
  });

});
