#[macro_use]
extern crate serde_derive;
extern crate chrono;
extern crate copy_dir;
extern crate docopt;
extern crate env_logger;
#[macro_use]
extern crate serde_json;
#[macro_use]
extern crate log;
extern crate actix_files;
extern crate actix_web;
extern crate colored;
extern crate tera;
extern crate toml;

extern crate rookeries;

use actix_web::middleware::Logger;
use actix_web::{App, HttpServer};
use chrono::prelude::*;
use colored::Colorize;
use docopt::Docopt;
use env_logger::{Builder, Env};
use rookeries::{Page, Site};
use serde_json::Value;
use std::collections::HashMap;
use std::ffi::OsString;
use std::fs::{create_dir, read_dir, read_to_string, remove_dir_all, write};
use std::path::{Path, PathBuf};

const USAGE: &'static str = "

Usage:
    rookeries (-h | --help)
    rookeries --version
    rookeries init [<project>]
    rookeries build [<project>]
    rookeries serve [<project>]

Options:
    -h --help       Show this help screen
    --version       Show version

Commands:
    init [<project>]    Initialize a project in the selected directory [defaults: .].
    build [<project>]   Builds the project into a static site [defaults: .].
    serve [<project>]   Serves the static site built from the project [defaults: .].
";

const ROOKERIES_HOME_ENV: &'static str = "ROOKERIES_HOME";
const ROOKERIES_HOME_DEFAULT_PATH: &'static str = "/usr/share/rookeries/";

#[derive(Debug, Deserialize)]
struct Args {
    flag_version: bool,
    arg_project: Option<String>,
    cmd_build: bool,
    cmd_init: bool,
    cmd_serve: bool,
}

fn initialize_logging() {
    // TODO: Add option to have --quiet, --verbose, --debug options for the logging.
    let env =
        Env::default().filter_or("ROOKERIES_LOG_LEVEL", "error,rookeries=info,actix_web=info");

    let mut builder = Builder::from_env(env);
    // TODO: Use the default format when supplying the --debug flag.
    builder
        .default_format_level(false)
        .default_format_timestamp(false)
        .default_format_module_path(false);
    builder.init();
}

fn get_project_directory(args: &Args) -> PathBuf {
    // TODO: Make more robust against errors and absences of the project directory.
    let project_directory = match &args.arg_project {
        None => String::from("."),
        Some(directory) => directory.clone(),
    };

    let project_directory = Path::new(&project_directory);
    project_directory.canonicalize().unwrap()
}

fn copy_file(
    source_directory: &PathBuf,
    target_directory: &PathBuf,
    filename: &str,
    file_type: &str,
) {
    match std::fs::copy(
        source_directory.join(filename),
        target_directory.join(filename),
    ) {
        Err(e) => {
            error!(
                "{}  Could not copy over {} {} because of: {}",
                "✘".red(),
                file_type,
                filename.red(),
                e
            );
            std::process::exit(2);
        }
        Ok(_) => {
            info!(
                "{}  Copied over {}: {}",
                "✔".green(),
                file_type,
                filename.green()
            );
        }
    }
}

fn main() {
    initialize_logging();

    let args: Args = Docopt::new(USAGE)
        .and_then(|d| d.deserialize())
        .unwrap_or_else(|e| e.exit());

    if args.flag_version {
        println!("{}", get_app_version());
        std::process::exit(0);
    }

    if args.cmd_serve {
        // TODO: Add an option to open up a browser window using xdg-open.

        // TODO: Detect and re-build project.
        // TODO: Create a file watcher to check for changes to the source files.
        let project_directory = get_project_directory(&args);

        // Check if build directory is available.
        let build_directory = project_directory.clone().join("build");
        if !build_directory.exists() || !build_directory.is_dir() {
            error!(
                "{}  This project needs a build!  Please run \"rookeries build {} \" and retry!",
                "✘".red(),
                &project_directory.display(),
            );
            std::process::exit(2);
        }

        info!("{}  Found a built site to serve!", "✔".green(),);

        let port = "8000";
        info!("");
        info!("Serving project at local port {}", port.cyan());

        HttpServer::new(move || {
            App::new()
                // TODO: Configure to serve the various files JS files loaded dynamically due to import() in Rookeries web app.
                .service(
                    actix_files::Files::new("/", build_directory.clone()).index_file("index.html"),
                )
                .wrap(Logger::default())
        })
        .bind(format!("127.0.0.1:{}", &port)) // TODO: Allow for finding a configuring a port.
        .unwrap()
        .run()
        .unwrap();
    }

    if args.cmd_build {
        let project_directory = get_project_directory(&args);

        // Check if project manifest is available.
        let project_manifest = project_directory.clone().join("site.toml");
        if project_manifest.exists() != true {
            error!(
                "{}  This project is missing a site.toml!  Please create one and retry!",
                "✘".red()
            );
            std::process::exit(2);
        }

        let project_manifest_details = read_to_string(project_manifest).unwrap();
        let project_site: Site = toml::from_str(&project_manifest_details).unwrap();
        info!(
            "{}  Found a project manifest for the \"{}\" site!",
            "✔".green(),
            &project_site.name
        );

        // Build a list of the pages to build out.
        info!("");
        info!("Searching for pages...");
        let mut site_source_pages: Vec<OsString> = Vec::new();
        for entry in read_dir(&project_directory.as_path()).unwrap() {
            // TODO: Make the search recursive through directories.
            let path = entry.ok().unwrap().path();
            if path.is_file() && path.extension() != None && path.extension().unwrap() == "md" {
                info!("{}  Found {}", "✔".green(), path.display());
                site_source_pages.push(path.into_os_string());
            }
        }

        // TODO: Add in support for figuring out the index page or allowing one to be set.
        info!("");
        info!("Compiling page information...");
        let page_overrides = project_site.clone().pages.unwrap_or(HashMap::new());

        // TODO: Add warning about a missing index page.
        // TODO: Add support for aliases (e.g. such as about being an index page)
        let site_source_pages: Vec<Page> = site_source_pages
            .iter()
            .map(|source_page| {
                let source_path = Path::new(source_page);
                let slug = String::from(source_path.file_stem().unwrap().to_str().unwrap());
                let content = read_to_string(source_path).unwrap();

                // TODO: Be able to extract the title from the header of the markdown content.
                let title = match &page_overrides.get(&slug) {
                    Some(page_entry) => {
                        let page_title = match page_entry.get("title") {
                            Some(title_) => {
                                let title_str = serde_json::to_string(title_).unwrap();
                                title_str.replacen("\"", "", 2)
                            }
                            None => format!("Rookeries :: {}", &slug),
                        };
                        page_title
                    }
                    None => format!("Rookeries :: {}", &slug).to_string(),
                };

                info!("{}  Compiling \"{}\" page", "✔".green(), &slug.green());

                Page {
                    title,
                    slug,
                    content,
                    created_at: None,
                }
            })
            .collect();

        // Create a build directory for the resulting files.
        info!("");
        info!("Preparing build directory...");
        let build_directory = project_directory.clone().join("build");
        if build_directory.exists() && build_directory.is_dir() {
            info!("{}  Recreating build directory...", "♼".yellow());
            remove_dir_all(&build_directory).expect("Failure!");
            create_dir(&build_directory).expect("Failure!");
            info!("{}  Created a build directory.", "✔".green());
        } else if !build_directory.exists() {
            create_dir(&build_directory).expect("Failure!");
            info!("{}  Created a build directory.", "✔".green());
        } else {
            error!("{}  Can not create a build directory!!!", "✘".red());
            std::process::exit(2);
        }

        // TODO: Create the html setup for the pages.
        info!("");
        info!("Initialize page rendering system...");
        let mut render_engine = tera::Tera::default();
        let template_path = &project_directory.join("template/base_index.html");
        render_engine
            .add_template_file(&template_path, Some("index"))
            .expect("Template failed to add");

        let status = json!({
            "version": env!("CARGO_PKG_VERSION").to_string(),
            "app": env!("CARGO_PKG_NAME").to_string(),
            "gitRevision": env!("GIT_REVISION").to_string(),
        });

        info!("");
        info!("Creating the index page...");
        let plugin_list = vec!["hello-world", "dark-mode-switch"];
        let mut ctx = tera::Context::new();
        ctx.insert("site", &project_site);
        // TODO: Ensure getting the right page as documented in the landingPage
        ctx.insert("current_page", "index");
        ctx.insert("plugins", &plugin_list);
        ctx.insert("rookeries", &status);
        let render_page_str = render_engine.render("index", &ctx).unwrap();

        write(build_directory.join("index.html"), render_page_str)
            .expect("Failed to write index.html");
        info!("{}  Created the index HTML page.", "✔".green());

        info!("");
        info!("Creating the pages...");
        for page in &site_source_pages {
            create_dir(&build_directory.join(&page.slug)).expect("Page directory creation failed.");
            ctx.insert("current_page", &page.slug);
            let render_page_str = render_engine.render("index", &ctx).unwrap();
            write(
                build_directory.join(&page.slug).join("index.html"),
                render_page_str,
            )
            .expect("Failed to write index.html");
            info!(
                "{}  Created the \"{}\" HTML page.",
                "✔".green(),
                &page.slug.green()
            );
        }

        // Create the API representations.
        info!("");
        info!("Creating the API JSON files...");
        let api_path = &build_directory.join("api");
        create_dir(&api_path).expect("Failed to create API directory");

        let site_json = serde_json::to_string_pretty(&project_site).unwrap();
        write(api_path.join("site.json"), site_json).expect("Failed to write site.json");
        info!(
            "{}  Created the API representation of the site.",
            "✔".green()
        );

        // Create a status.json that is really a build metadata artifact.
        // TODO: Clean up status setup.
        let status = serde_json::to_string_pretty(&status).unwrap();
        write(api_path.join("_build.json"), status).expect("Failed to write the build info.");
        info!(
            "{}  Created the API representation of the status.",
            "✔".green()
        );

        // Create all the page API reps.
        let api_pages_path = &api_path.join("pages");
        create_dir(&api_pages_path).expect("Failed to create API for pages directory");
        for page in site_source_pages {
            let page_json = serde_json::to_string_pretty(&page).unwrap();
            let page_json_filename = format!("{}.json", &page.slug);
            write(api_pages_path.join(&page_json_filename), page_json)
                .expect("Failed to write the page json");
            info!(
                "{}  Created the API representation of the \"{}\" page.",
                "✔".green(),
                &page.slug.green()
            );
        }

        // Bring in the static file directory.
        // TODO: Ensure a separate copy over of JS assets from Rookeries.  (from plugins and standalone compiled JS)
        let assets_source_dir = Path::new(&project_directory)
            .join("static")
            .canonicalize()
            .unwrap();
        let assets_target_dir = Path::new(&build_directory).join("static");
        info!("");
        info!(
            "Copying static assets from \"{}\"  to \"{}\"",
            &assets_source_dir.display().to_string().green(),
            &assets_target_dir.display().to_string().green()
        );
        match copy_dir::copy_dir(&assets_source_dir, &assets_target_dir) {
            Err(e) => {
                error!(
                    "{}  Could not copy over the static assets because of: {}",
                    "✘".red(),
                    e
                );
                std::process::exit(2);
            }
            Ok(v) => {
                if v.len() > 0 {
                    // TODO: Improve this...
                    error!("{}  Could not copy over the static assets.", "✘".red());
                    std::process::exit(2);
                } else {
                    info!("{}  Copied over assets!", "✔".green());
                }
            }
        }

        // TODO: Remove the copied over base_index.html

        // Copy over plugins too!
        info!("");
        info!("Activate plugins...");
        // TODO: Copy over only active and available markdown plugins.
        let plugin_list = vec!["hello-world", "dark-mode-switch"];
        let plugins_install_path = Path::new(&project_directory)
            .join("plugins")
            .canonicalize()
            .unwrap();

        let static_js_path = Path::new(&assets_target_dir).join("js");
        for plugin in plugin_list {
            let plugin_file_name = format!("{}-plugin.js", plugin);
            copy_file(
                &plugins_install_path,
                &static_js_path,
                &plugin_file_name,
                "plugin",
            );
        }

        info!(
            "{}  You should be good to go!  {}",
            "✔".green(),
            "✔".green()
        );
    }

    if args.cmd_init {
        info!("");
        info!("Initializing a new site...");

        // TODO: Add support for an override flag.
        // Build out the project directory.
        let project_directory = match &args.arg_project {
            None => String::from("."),
            Some(directory) => directory.clone(),
        };

        let project_directory = Path::new(&project_directory);
        if !project_directory.exists() {
            create_dir(&project_directory).expect("Could not create the project directory.");
            let project_dir = &project_directory.canonicalize().unwrap();
            info!(
                "{}  Created a new project in {}",
                "✔".green(),
                format!("{}", &project_dir.display()).green()
            );
        }
        let project_directory = project_directory
            .canonicalize()
            .expect("Project directory fails.");

        // TODO: Make the sample pages happen.
        let mut sample_pages: HashMap<String, Value> = HashMap::new();
        sample_pages.insert("index".to_string(), json!({"title": "Welcome"}));
        sample_pages.insert("demo".to_string(), json!({"title": "Demo"}));
        sample_pages.insert(
            "license".to_string(),
            json!({"title": "License (Apache v2.0)"}),
        );

        // Create a site toml if one is missing.
        let project_manifest = project_directory.join("site.toml");
        match &project_manifest.exists() {
            true => {
                info!("{}  Found existing site.toml.  Skipping.", "♼".yellow());
            }
            false => {
                // TODO: Support an interactive setup of the Rookeries site.
                let mut site = Site::default();
                site.name = "New Rookeries Site".to_string();
                site.pages = Some(sample_pages.clone());

                let site_toml_content = toml::to_string(&site).unwrap();

                // TODO: Consider a more complex template for the site.
                let rookeries_version = env!("CARGO_PKG_VERSION");
                let generation_timestamp: DateTime<Local> = Local::now();
                let generation_timestamp =
                    generation_timestamp.format("%B %e, %Y @ %R").to_string();
                let site_toml_header = format!(
                    "# Site generated by Rookeries v{} on {}",
                    rookeries_version, generation_timestamp
                );

                let site_toml = format!("{}\n\n{}", site_toml_header, site_toml_content);

                write(&project_manifest, site_toml).expect("Failed to write site.toml");
                info!("{}  Created a site.toml!", "✔".green());
            }
        }

        let rookeries_home_path = match std::env::var_os(&ROOKERIES_HOME_ENV) {
            None => Path::new(&ROOKERIES_HOME_DEFAULT_PATH)
                .canonicalize()
                .unwrap(),
            Some(path) => Path::new(&path).canonicalize().unwrap(),
        };

        // TODO: Start setup of plugins and templates
        // Template copy over and setup
        info!("");
        info!("Add templates to site");
        let template_path = &rookeries_home_path.join("template");
        let site_template_path = &project_directory.join("template");
        if site_template_path.exists() {
            // TODO: Figure out override.
            info!("{}  Found existing templates.  Skipping.", "♼".yellow());
        } else {
            create_dir(&site_template_path).expect("Failed to create template directory");
            // TODO: Allow for more files.
            copy_file(
                &template_path,
                &site_template_path,
                "base_index.html",
                "template HTML",
            );
        }

        info!("");
        info!("Add template static assets to site");
        let site_static_path = &project_directory.join("static");
        if site_static_path.exists() {
            info!(
                "{}  Found existing static assets.  Skipping.",
                "♼".yellow()
            );
        } else {
            match copy_dir::copy_dir(&template_path, &site_static_path) {
                Err(e) => {
                    error!(
                        "{}  Could not copy over the template static assets because of: {}",
                        "✘".red(),
                        e
                    );
                    std::process::exit(2);
                }
                Ok(v) => {
                    if v.len() > 0 {
                        // TODO: Improve this...
                        error!(
                            "{}  Could not copy over the template static assets.",
                            "✘".red()
                        );
                        std::process::exit(2);
                    } else {
                        info!("{}  Copied over the template assets!", "✔".green());
                    }
                }
            }
        }

        info!("");
        info!("Add sample markdown files to site");
        for page in sample_pages.keys() {
            let page_file_name = format!("{}.md", &page);
            let sample_page_path = project_directory.join(&page_file_name);
            match &sample_page_path.exists() {
                true => {
                    info!(
                        "{}  Found existing {}.  Skipping.",
                        "♼".yellow(),
                        &page_file_name.yellow()
                    );
                }
                false => {
                    copy_file(
                        &template_path,
                        &project_directory,
                        &page_file_name,
                        "sample markdown",
                    );
                }
            }
        }

        // Do the same thing for plugins
        // Template copy over and setup
        info!("");
        info!("Add plugins to site");
        let plugins_path = &rookeries_home_path.join("plugins");
        let site_plugins_path = &project_directory.join("plugins");
        if site_plugins_path.exists() {
            // TODO: Figure out override.
            info!("{}  Found existing plugins.  Skipping.", "♼".yellow());
        } else {
            create_dir(&site_plugins_path).expect("Failed to create plugin directory");
            let plugin_list = vec!["hello-world", "dark-mode-switch"];
            for plugin in plugin_list {
                let plugin_file_name = format!("{}-plugin.js", &plugin);
                copy_file(
                    &plugins_path,
                    &site_plugins_path,
                    &plugin_file_name,
                    "plugin",
                );
                info!(
                    "{}  Copied over the {} plugin!",
                    "✔".green(),
                    &plugin.green()
                );
            }
        }
    }
}

fn get_app_version() -> String {
    format!(
        "rookeries v{version} © 2013-2019 {authors}",
        version = env!("CARGO_PKG_VERSION"),
        authors = env!("CARGO_PKG_AUTHORS"),
    )
    .to_string()
}
