/*
 Rookeries - Entrypoint for browser app

 @copyright (c) Copyright 2013-2019 Dorian Pula
 @license Apache 2.0
 @author Dorian Pula [dorian.pula@amber-penguin-software.ca]
 */

import markdown from "markdown-it";
import highlighter from "./syntax-highlighting";  // eslint-disable-line

const UPDATE_ROUTE_EVENT = "rookeries-router-push";
const PAGE_CONTENT_FETCHED_EVENT = "rookeries-content-fetched";
// Use a default of 5 hours in milliseconds.
const DEFAULT_CACHE_LIMIT = 5 * 60 * 1000;


export class RookeriesApp extends HTMLElement {
  static get observedAttributes() {
    return ["title", "cache-limit"];
  }

  static get tagName() {
    return "rookeries-app"
  }

  static storageKey(pageSlug) {
    return `rookeries::${pageSlug}`;
  }

  connectedCallback() {
    this.render();
    window.addEventListener(UPDATE_ROUTE_EVENT, RookeriesApp.triggerRender);
    window.addEventListener(PAGE_CONTENT_FETCHED_EVENT, RookeriesApp.triggerRender);
    window.addEventListener("popstate", RookeriesApp.triggerRender);
  }

  static triggerRender() {
    const element = window.document.querySelector(RookeriesApp.tagName);
    element.render();
  }

  static isLocalLink(url) {
    return url.startsWith("/") && !url.startsWith("/static");
  }

  static pushUrl(event) {
    if (event.type !== "click" && event.type !== "keydown" && event.code !== "Enter") {
      return;
    }

    const targetUrl = event.currentTarget.getAttribute("href");
    if (!RookeriesApp.isLocalLink(targetUrl)) {
      return;
    }
    event.preventDefault();
    window.history.pushState({}, "Rookeries", targetUrl);
    window.dispatchEvent(new Event(UPDATE_ROUTE_EVENT));
  }

  async fetchPageContent(pageSlug) {
    const response = await fetch(`/api/pages/${pageSlug}.json`);
    const json = await response.json();

    const content = Reflect.get(json, "content") || "";
    const title = Reflect.get(json, "title") || "";
    const fetchedOn = new Date().toJSON();

    const storedContent = JSON.stringify({content, title, fetchedOn});

    window.localStorage.setItem(RookeriesApp.storageKey(pageSlug), storedContent);
  }

  render() {
    const currentPath = new URL(window.location).pathname;
    let currentPage = currentPath === "/" ? "index" : currentPath.slice(1);
    if (currentPage.endsWith("/")) {
      currentPage = currentPage.slice(0, currentPage.length - 1);
    }

    const rawStoredContent = window.localStorage.getItem(RookeriesApp.storageKey(currentPage));
    const storedContent = rawStoredContent !== null ? JSON.parse(rawStoredContent) : null;
    const storedContentFetchTime = storedContent !== null && storedContent.fetchedOn !== undefined
      ? new Date(storedContent.fetchedOn)
      : new Date();
    const elapsedTime = new Date() - storedContentFetchTime;
    const cachedTimeDiff = this.getAttribute("cache-limit") || DEFAULT_CACHE_LIMIT;

    if (!storedContent || (storedContent && elapsedTime > cachedTimeDiff)) {
      this.fetchPageContent(currentPage).then(() => {
        window.dispatchEvent(new Event(PAGE_CONTENT_FETCHED_EVENT));
      });
    }

    const md = markdown({
      html: true,
      highlight: (code_) => {
        return highlighter.highlightAuto(code_).value;
      }
    });

    const content = storedContent !== null && storedContent.content !== undefined
      ? storedContent.content
      : `<h1>Loading...</h1>`;
    const renderedContent = md.render(content);
    this.innerHTML = `<article>${renderedContent}</article>`;

    const siteTitle = this.getAttribute("title") || "Rookeries";
    const pageTitle = storedContent !== null && storedContent.title !== undefined
      ? storedContent.title
      : currentPage;
    window.document.title = `${pageTitle} - ${siteTitle}`;

    const hyperlinks = document.querySelectorAll("a");
    hyperlinks.forEach((link) => {
      link.addEventListener("click", RookeriesApp.pushUrl);
      link.addEventListener("keydown", RookeriesApp.pushUrl);
    });
  }
}

if (customElements.get(RookeriesApp.tagName) === undefined) {
  customElements.define(RookeriesApp.tagName, RookeriesApp);
}
